# Generated by Django 4.2.2 on 2023-10-08 20:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0015_alter_annotationfield_vocabulary"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="annotationvalue",
            constraint=models.UniqueConstraint(
                fields=("entity", "field"), name="uniquetogether_entity_field"
            ),
        ),
    ]
