# Generated by Django 4.2.7 on 2024-01-09 07:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0020_alter_annotationvalue_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationvalue",
            name="modified",
            field=models.DateTimeField(auto_now=True),
        ),
    ]
